"""geographiclib: geodesic routines from GeographicLib"""

__version_info__ = (2,
                    1,
                    0)
"""GeographicLib version as a tuple"""

__version__ = "2.1"
"""GeographicLib version as a string"""

__release_date__ = "2025-08-21"
"""GeographicLib release date"""
